/*
 * ValueHolder.hpp
 *
 *  Created on: 11. 7. 2017
 *	  Author: Jan Travnicek
 */

#ifndef _VALUE_HOLDER_HPP_
#define _VALUE_HOLDER_HPP_

#include <alib/memory>
#include <alib/typeindex>
#include <alib/typeinfo>
#include <abstraction/ValueHolderInterface.hpp>

namespace abstraction {

template < class Type >
class ValueHolderImpl : public ValueHolderInterface < Type > {
	mutable std::optional < Type > m_data;

public:
	using ValueHolderInterface < Type >::ValueHolderInterface;

	void setData ( Type && data ) {
		m_data = std::move ( data );
	}

	Type && getValue ( ) const override {
		return std::move ( m_data.value ( ) );
	}
};

template < class Type >
class ValueHolderImpl < const Type > : public ValueHolderInterface < Type > {
	mutable std::optional < Type > m_data;

public:
	using ValueHolderInterface < Type >::ValueHolderInterface;

	void setData ( const Type && data ) {
		m_data = std::move ( const_cast < Type && > ( data ) );
	}

	Type && getValue ( ) const override {
		return std::move ( m_data.value ( ) );
	}
};

template < class Type >
class ValueHolderImpl < Type & > : public ValueHolderInterface < Type > {
	mutable std::optional < std::reference_wrapper < Type > > m_data;

public:
	using ValueHolderInterface < Type >::ValueHolderInterface;

	void setData ( Type & data ) {
		m_data = std::reference_wrapper < Type > ( data );
	}

	Type && getValue ( ) const override {
		return std::move ( m_data->get ( ) );
	}
};

template < class Type >
class ValueHolderImpl < const Type & > : public ValueHolderInterface < Type > {
	mutable std::optional < std::reference_wrapper < Type > > m_data;

public:
	using ValueHolderInterface < Type >::ValueHolderInterface;

	void setData ( const Type & data ) {
		m_data = std::reference_wrapper < Type > ( const_cast < Type & > ( data ) );
	}

	Type && getValue ( ) const override {
		return std::move ( m_data->get ( ) );
	}
};

template < class Type >
class ValueHolderImpl < Type && > : public ValueHolderInterface < Type > {
	mutable std::optional < std::reference_wrapper < Type > > m_data;

public:
	using ValueHolderInterface < Type >::ValueHolderInterface;

	void setData ( Type && data ) {
		m_data = std::reference_wrapper < Type > ( data );
	}

	Type && getValue ( ) const override {
		return std::move ( m_data->get ( ) );
	}
};

template < class Type >
class ValueHolderImpl < const Type && > : public ValueHolderInterface < Type > {
	mutable std::optional < std::reference_wrapper < Type > > m_data;

public:
	using ValueHolderInterface < Type >::ValueHolderInterface;

	void setData ( const Type && data ) {
		m_data = std::reference_wrapper < Type > ( const_cast < Type & > ( data ) );
	}

	Type && getValue ( ) const override {
		return std::move ( m_data->get ( ) );
	}
};

template < class Type >
class ValueHolder : public ValueHolderImpl < Type > {
public:
	ValueHolder ( const std::shared_ptr < abstraction::OperationAbstraction > & ref, Type && value, bool temporary ) : ValueHolderImpl < Type > ( ref, ParamQualifiers::paramQualifiers < Type > ( ), temporary ) {
		this->setData ( std::forward < Type > ( value ) );
	}

	std::shared_ptr < abstraction::Value > clone ( const std::shared_ptr < abstraction::OperationAbstraction > & lifeReference, ParamQualifiers::ParamQualifierSet paramQualifierSet, bool isTemporary, bool move ) override {
		if ( ( paramQualifierSet && ParamQualifiers::ParamQualifierSet::LREF || paramQualifierSet && ParamQualifiers::ParamQualifierSet::RREF ) && this->isTemporary ( ) && ! this->isRef ( ) && lifeReference == nullptr )
			throw std::domain_error ( "Taking reference to a temporary" );

		if ( ( paramQualifierSet && ParamQualifiers::ParamQualifierSet::CONST ) && ( paramQualifierSet && ParamQualifiers::ParamQualifierSet::LREF ) )
			return std::make_shared < abstraction::ValueHolder < const std::decay_t < Type > & > > ( lifeReference, retrieveValue < const std::decay_t < Type > & > ( this->getProxyAbstraction ( ), move ), isTemporary );
		else if ( ( paramQualifierSet && ParamQualifiers::ParamQualifierSet::CONST ) && ( paramQualifierSet && ParamQualifiers::ParamQualifierSet::RREF ) )
			return std::make_shared < abstraction::ValueHolder < const std::decay_t < Type > && > > ( lifeReference, retrieveValue < const std::decay_t < Type > && > ( this->getProxyAbstraction ( ), move ), isTemporary );
		else if ( paramQualifierSet && ParamQualifiers::ParamQualifierSet::LREF )
			return std::make_shared < abstraction::ValueHolder < std::decay_t < Type > & > > ( lifeReference, retrieveValue < std::decay_t < Type > & > ( this->getProxyAbstraction ( ), move ), isTemporary );
		else if ( paramQualifierSet && ParamQualifiers::ParamQualifierSet::RREF )
			return std::make_shared < abstraction::ValueHolder < std::decay_t < Type > && > > ( lifeReference, retrieveValue < std::decay_t < Type > && > ( this->getProxyAbstraction ( ), move ), isTemporary );
		else if constexpr ( std::is_abstract_v < std::decay_t < Type > > )
			throw std::domain_error ( "Cannot declare value of abstract class." );
		else if constexpr ( ! std::is_assignable_v < std::decay_t < Type > &, std::decay_t < Type > > )
			throw std::domain_error ( "Cannot assign value." );
		else if ( paramQualifierSet && ParamQualifiers::ParamQualifierSet::CONST )
			return std::make_shared < abstraction::ValueHolder < const std::decay_t < Type > > > ( lifeReference, retrieveValue < const std::decay_t < Type > > ( this->getProxyAbstraction ( ), move ), isTemporary );
		else
			return std::make_shared < abstraction::ValueHolder < std::decay_t < Type > > > ( lifeReference, retrieveValue < std::decay_t < Type > > ( this->getProxyAbstraction ( ), move ), isTemporary );
	}

	abstraction::ParamQualifiers::ParamQualifierSet getTypeQualifiers ( ) const override {
		return abstraction::ParamQualifiers::paramQualifiers < Type > ( );
	}

	ext::type_index getTypeIndex ( ) const override {
		return ext::type_index ( typeid ( std::decay_t < Type > ) );
	}

};

} /* namespace abstraction */

#endif /* _VALUE_HOLDER_HPP_ */
