#ifndef _CLI_PRINT_COMMAND_H_
#define _CLI_PRINT_COMMAND_H_

#include <ast/Command.h>
#include <environment/Environment.h>
#include <ast/Statement.h>

#include <global/GlobalData.h>
#include <registry/Registry.h>

namespace cli {

class PrintCommand : public Command {
	std::shared_ptr < Statement > m_command;

public:
	PrintCommand ( std::shared_ptr < Statement > command ) : m_command ( std::move ( command ) ) {
	}

	CommandResult run ( Environment & environment ) const override {
		std::shared_ptr < abstraction::Value > value = m_command->translateAndEval ( nullptr, environment );

		std::shared_ptr < abstraction::OperationAbstraction > res = abstraction::Registry::getValuePrinterAbstraction ( value->getType ( ) );

		res->attachInput ( value, 0, false );
		res->attachInput ( std::make_shared < abstraction::ValueHolder < std::ostream & > > ( nullptr, common::Streams::out, false ), 1, false );
		std::shared_ptr < abstraction::Value > result = res->eval ( );
		if ( ! result )
			throw std::invalid_argument ( "Eval of result print statement failed." );

		return CommandResult::OK;
	}
};

} /* namespace cli */

#endif /* _CLI_PRINT_COMMAND_H_ */
