#ifndef _PARAM_QUALIFIERS_HPP_
#define _PARAM_QUALIFIERS_HPP_

#include <alib/set>
#include <alib/type_traits>

namespace abstraction {

class ParamQualifiers {
public:
	enum class ParamQualifierSet {
		NONE = 0x0,
		CONST = 0x1,
		LREF = 0x2,
		RREF = 0x4,
	};

	friend constexpr ParamQualifierSet operator | ( ParamQualifierSet first, ParamQualifierSet second ) {
		return static_cast < ParamQualifierSet > ( static_cast < unsigned > ( first ) | static_cast < unsigned > ( second ) );
	}

	friend constexpr ParamQualifierSet operator & ( ParamQualifierSet first, ParamQualifierSet second ) {
		return static_cast < ParamQualifierSet > ( static_cast < unsigned > ( first ) & static_cast < unsigned > ( second ) );
	}

	friend constexpr bool operator && ( ParamQualifierSet first, ParamQualifierSet second ) {
		return ( static_cast < unsigned > ( first ) & static_cast < unsigned > ( second ) ) == static_cast < unsigned > ( second );
	}

	template < class Param >
	static constexpr ParamQualifierSet paramQualifiers ( ) {
		ParamQualifierSet res = ParamQualifierSet::NONE;

		if ( std::is_lvalue_reference < Param >::value )
			res = res | ParamQualifierSet::LREF;

		if ( std::is_rvalue_reference < Param >::value )
			res = res | ParamQualifierSet::RREF;

		if ( std::is_const < typename std::remove_reference < Param >::type >::value )
			res = res | ParamQualifierSet::CONST;

		return res;
	}

	friend std::ostream & operator << ( std::ostream & os, ParamQualifierSet paramQualifierSet );
};

} /* namespace abstraction */

#endif // _PARAM_QUALIFIERS_HPP_
